#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "xmalloc.h"
#include "util.h"
#include "object.h"
#include "classmember.h"
#include "class.h"
#include "directory.h"
#include "source.h"
#include "collection.h"
#include "variable.h"
#include "function.h"

#define basename FUDGE_BASENAME
#include "demangle.h"
#undef basename

Cvariable		*Cvariable::g_hash[HASHSIZE] = {0};
unsigned int	Cvariable::g_hash_cnt        = 0;
Cvariable		*Cvariable::g_local_headP    = 0;

/*
void
trap_variable(void)
{
	fprintf(stderr,"** Trapped\n");
}
*/

Cvariable::Cvariable(const char *nameP) : Cclassmember(nameP)
{
	m_hashP       = 0;
	m_local_nextP = 0;
}

void
Cvariable::reset(void)
{
	int 		i;
	Cvariable	*variableP;

	for (i = 0; i < HASHSIZE; ++i) {
		for (variableP = g_hash[i]; variableP; variableP = variableP->m_hashP) {
			variableP->m_flags |= InvisibleX;
}	}	}

Cvariable *
Cvariable::locate(const char *nameP, int flags, Centity *sourceP, Centity *functionP)
{
	unsigned int	index;
	Cvariable		**variablePP, *variableP, *bestP;
	Centity			*parentP, *grandparentP;
	Cfunction		*function1P;
	const char		*P;

/*
	if (!strcmp(nameP, "__tcf_1") && !strcmp(sourceP->m_nameP, "as_license.cpp")) {
		trap_variable();
	}
 */
	bestP = 0;
	index = hash(nameP);
	for (variablePP = g_hash + index; (variableP = *variablePP); variablePP = &(variableP->m_hashP)) {
		if (strcmp(variableP->m_nameP, nameP)) {
			// Name differs so not the right variable
			continue;
		}
		if (variableP->m_flags & InvisibleX) {
			continue;
		}
		parentP = variableP->m_parentP;		// null if don't know parent
		if (variableP->m_flags & FunctionX) {
			if (flags & AddressesX) {
				grandparentP = parentP->m_parentP;
				if (grandparentP == sourceP) {
					return(variableP);
			}	}
			continue;
		}

		if (!(variableP->m_flags & NestedX)) {
			// If it is Nested it has some unique .xxxx number on end
			if (flags & LocalX) {
				// Know this should be in our source
				if (parentP != sourceP) {
					continue;
			}	}
		} else {
			if (parentP && parentP != functionP) {
				continue;
		}	}

		if (!bestP) {
			bestP = variableP;
			continue;
		}
		if (!(bestP->m_flags & NestedX)) {
			if (parentP == sourceP || parentP == functionP) {
				bestP = variableP;
	}	}	}

	if (!bestP) {
		bestP = new Cvariable(nameP);
		if (!bestP) {
			outofmemory();
		}
		/* Add to end so always return first function declaration seen if
		 * searching for location of the function declaration:
		 *
	     * eg. if we are looking for foo and foo was in foo.c later
	     *     also seen in foo.o return reference to foo.c:foo in
		 *     preferance to foo.o:foo
	     */
	
		*variablePP = bestP;
		++Cvariable::g_hash_cnt;

		bestP->m_flags   = flags | ExternalX;
		bestP->m_parentP = sourceP;

		if (flags & AddressesX) {
			function1P = Cfunction::local_exists(nameP, sourceP);
			if (function1P) {
				bestP->m_parentP = function1P;
				bestP->m_flags  &= ~ExternalX;
				bestP->m_flags  |= FunctionX;
				bestP->add_created();
				return(bestP);
		}	} 
		if (flags & LocalX) {
			P = strchr(nameP, '.');
			if (P != 0) {
				/* The assembler labels local static variables declared within
				 * functions with a numeric suffix to avoid collisions with
				 * global variables and the same named local variables in other
				 * functions.
			 	 */
				bestP->m_flags  |= NestedX;
				flags           |= NestedX;
		}	}
		if (functionP) {
			if (flags & (ParameterX|DynamicX|NestedX)) {
				bestP->m_parentP = functionP;
	}	}	}
	if (bestP->m_flags & ExternalX) {
		if (flags & FindX) {
			if (!(flags & AddressesX)) {
				bestP->m_flags &= ~AddressesX;
			}
		} else {
			if (bestP->m_flags & NestedX) {
				bestP->m_local_nextP = g_local_headP;
				g_local_headP        = bestP;
			} else {
				bestP->m_parentP = sourceP;
			}	
			bestP->m_flags |= flags;
			bestP->m_flags &= ~(AddressesX | ExternalX);
			bestP->add_created();
	}	}
	return(bestP);
}

void
Cvariable::resolve(int show_variables, int no_function_calls)
{
	Csource		*sourceP;
	Cfunction	*functionP;
	Cvariable	*variableP;
	char		*nameP, *P;
	int 		i, flags;
	
	sourceP = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (variableP = g_hash[i]; variableP; variableP = variableP->m_hashP) {
			nameP = variableP->m_nameP;
			flags = variableP->m_flags;
			if (!(flags & UsedX)) {
				variableP->m_flags |= IgnoreX;
				continue;
			}
			if (flags & ExternalX) {
				variableP->add_created();
				if (flags & AddressesX) {
					functionP = Cfunction::global_exists(nameP);
					if (functionP) {
						if (no_function_calls) {
							variableP->m_flags |= IgnoreX;
							continue;
						}
						variableP->m_parentP = functionP;
						variableP->m_flags  &= ~ExternalX;
						continue;
					}
					variableP->m_flags &= ~AddressesX;
				}
				if (!sourceP) {
					sourceP    = new Ccollection(":globals:", 0);
					if (!sourceP) {
						outofmemory();
					}
				}
				variableP->m_parentP = sourceP;
				continue;
			}

			if (flags & NestedX) {
				P = strchr(nameP, '.');
				if (P) {
					// Remove .<number> from the end of variable names
					*P = 0;
			}	}
			switch (*nameP) {
			case '_':
				if (!strcmp(nameP,"__PRETTY_FUNCTION__")) {
					variableP->m_flags |= IgnoreX;
					continue;
				}
				break;
			}

			switch (show_variables) {
			case Global_variables:
				if (flags & LocalX) {
					variableP->m_flags |= IgnoreX;
					continue;
				}
				break;
			case Static_variables:
				if (flags & NestedX) {
					variableP->m_flags |= IgnoreX;
					continue;
			}	}
	}	}
}

/* virtual */ objectE
Cvariable::type(void) const
{
	return(variableE);
}

/* virtual */ void
Cvariable::emit_id(void) const
{
	fprintf(stdout, "V%p", this);
}

const char *
Cvariable::get_type(void) const
{
	int			flags;
	const char 	*typeP;

	flags = m_flags;
	if (flags & ParameterX) {
		typeP = "PV";		// Parameter variable
	} else if (flags & DynamicX) {
		typeP = "LV";		
	} else if (flags & NestedX) {
		typeP = "FV";		// Function variable
	} else if (flags & LocalX) {
		typeP = "SV";		// Static variable
	} else if (flags & WeakX) {
		char	*nameP;
		typeP = "WV";
		nameP = m_nameP;
		if (!strncmp(nameP, "_ZT", 3)) {
			switch (nameP[3]) {
			case 'V':
				typeP = "TV";
				break;
			case 'I':
				typeP = "TI";
				break;
			case 'S':
				typeP = "TN";
				break;
		}	}
	} else if (flags & ExternalX) {
		typeP = "EV";
	} else if (flags & AddressesX) {
		typeP = "FA";
	} else {
		typeP = "GV";
	}
	return(typeP);
}

/* virtual */ void
Cvariable::emit_edges(void)
{
	if (!(m_flags & IgnoreX)) {
		Cclassmember::emit_edges();
		if (Cclass::g_headP) {
			emit_class();
	}	}
}
		
/* virtual */ void
Cvariable::emit_instance(void) const
{
	fprintf(stdout, "$INSTANCE V%p %s\n", this, get_type());
}

void
Cvariable::emit_attributes(void) const
{
	const char	*nameP;
	char		*demangleP;

	emit_id();
	fputs(" {", stdout);

	nameP     = m_nameP;
	demangleP = 0;
	if (nameP[0] == '_') {
		demangleP = cplus_demangle(nameP, DMGL_ANSI|DMGL_PARAMS|DMGL_GNU_V3);
		if (demangleP) {
			fprintf(stdout, "mangle=\"%s\" ", nameP);
			nameP = demangleP;
	}	}
	fprintf(stdout, "label=\"");
	if (m_flags & AddressesX) {
		fprintf(stdout, "&");
	}
	fprintf(stdout, "%s\" file=\".\"", nameP);
	if (demangleP) {
		free(demangleP);
	}
	fputs("}\n", stdout);
}

/* virtual */ void
Cvariable::print(void) const
{
	fputs(m_nameP, stdout);
}
